package org.openxma.dsl.dom.linking;

import java.io.File;
import java.net.URL;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.mwe.core.resources.ResourceLoaderFactory;
import org.eclipse.xtext.linking.impl.DefaultLinkingService;
import org.eclipse.xtext.linking.impl.IllegalNodeException;
import org.eclipse.xtext.parsetree.AbstractNode;
import org.eclipse.xtext.resource.ClasspathUriUtil;
import org.eclipse.xtext.resource.IEObjectDescription;
import org.eclipse.xtext.resource.IResourceDescription;
import org.eclipse.xtext.scoping.IScope;
import org.eclipse.xtext.scoping.Scopes;
import org.eclipse.xtext.scoping.impl.SimpleScope;
import org.eclipse.xtext.util.SimpleAttributeResolver;
import org.openxma.dsl.common.DslConstants;
import org.openxma.dsl.core.CorePackage;
import org.openxma.dsl.dom.DomPackage;

import com.google.common.collect.Iterators;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;

public class DomDslLinkingService extends DefaultLinkingService {

	private Map<String, List<URI>> uriCacheMap = Maps.newHashMap();

	public List<EObject> getLinkedObjects(EObject context, EReference ref,
			AbstractNode node) throws IllegalNodeException {

		final EClass requiredType = ref.getEReferenceType();
		if (requiredType == null) {
			return Collections.<EObject> emptyList();
		}
		final String referencedName = getCrossRefNodeAsString(node);
		if (referencedName != null) {
			IScope scope = IScope.NULLSCOPE;
			URI contextUri = context.eResource().getURI();
			if ((DomPackage.eINSTANCE.getEntity_SuperType().equals(ref) 
					|| CorePackage.eINSTANCE.getDataTypeAndTypeParameter_DataType().equals(ref)
			)
					&& DslConstants.OPENXMA_DOM_FILE_EXTENSION
							.equals(contextUri.fileExtension())) {
				List<URI> domDslPackageUris = getDomDslPackageUris(context,
						contextUri);
				Collections.sort(domDslPackageUris, new UriComparator(
						referencedName));
				scope = getPackageScope(getScope(context, ref), context, ref,
						domDslPackageUris);
			} else {
				scope = getScope(context, ref);
			}

			
			IEObjectDescription eObjectDescription = scope.getContentByName(referencedName);
			if (eObjectDescription != null) {
				return Collections.singletonList(eObjectDescription.getEObjectOrProxy());
			}
		}
		return Collections.emptyList();
	}

	public List<URI> getDomDslPackageUris(EObject context, URI contextUri) {
		List<URI> packageUris = uriCacheMap.get(contextUri.toString());
		if (null == packageUris) {
			packageUris = Lists.newArrayList();
			if (ClasspathUriUtil.isClasspathUri(contextUri)) {
				String pathAsString = contextUri.path();
				if (contextUri.hasAbsolutePath()) {
					pathAsString = pathAsString.substring(1);
				}
				URL contextUrl = ResourceLoaderFactory.createResourceLoader()
						.getResource(pathAsString);
				if (null == contextUrl) {
					contextUrl = Thread.currentThread().getContextClassLoader()
							.getResource(pathAsString);
				}
				File domFile = new File(contextUrl.getFile());
				for (File file : domFile.getParentFile().listFiles()) {
					if (!file.getName().equals(domFile.getName())
							&& file.getName().endsWith(
									DslConstants.OPENXMA_DOM_FILE_EXTENSION)) {
						URI classPathUri = contextUri.trimSegments(1)
								.appendSegment(file.getName());
						packageUris.add(classPathUri);
					}
				}
			}
			uriCacheMap.put(contextUri.toString(), packageUris);
		}
		return packageUris;
	}

	protected IScope getPackageScope(final IScope scope, EObject context,
			EReference reference, final List<URI> urisWithinPackage) {
		IScope lazyParentScope = IScope.NULLSCOPE;
		for (URI uri : urisWithinPackage) {
			lazyParentScope = newPackageScope(lazyParentScope, context, reference, uri);
		}
		return new SimpleScope(lazyParentScope, null) {
			@Override
			public IEObjectDescription getContentByName(String name) {
				IEObjectDescription objectDescription  = scope.getContentByName(name);
				return objectDescription !=null ? objectDescription : getOuterScope().getContentByName(name);
			}
		};
	}

	protected IScope newPackageScope(IScope parentScope, EObject context, EReference reference, URI uri) {
		return new SimpleScope(parentScope, new LazyResourceLoadingIterable(reference.getEReferenceType(), context, uri));
	}

	private final class UriComparator implements Comparator<URI> {
		private final String referencedName;

		private UriComparator(String referencedName) {
			this.referencedName = referencedName;
		}

		public int compare(URI uri1, URI uri2) {
			Integer u1 = compare(uri1, referencedName);
			Integer u2 = compare(uri2, referencedName);
			return u1.compareTo(u2);
		}

		private int compare(URI uri, final String referencedName) {
			return uri.lastSegment().toUpperCase().equals(
					referencedName.toUpperCase() + "."
							+ DslConstants.OPENXMA_DOM_FILE_EXTENSION) ? 1
					: (uri.lastSegment().toUpperCase().contains(
							referencedName.toUpperCase()) ? 0 : -1);
		}
	}

	public static class LazyResourceLoadingIterable implements Iterable<IEObjectDescription> {
		private IResourceDescription resourceDescription;
		private EClass eClass;
		private URI uri;
		private EObject context;

		public LazyResourceLoadingIterable(IResourceDescription resourceDescription, EClass eReferenceType) {
			this.resourceDescription = resourceDescription;
			this.eClass = eReferenceType;
		}
		
		public LazyResourceLoadingIterable( EClass eReferenceType, EObject context, URI uri) {
			this.eClass = eReferenceType;
			this.context = context;
			this.uri = uri;
		}

		public Iterator<IEObjectDescription> iterator() {
			if (uri != null ) {
				return Scopes
				.allInResource(context.eResource().getResourceSet().getResource(uri, true), eClass,
						SimpleAttributeResolver.NAME_RESOLVER).iterator() ;
			} 
			return resourceDescription==null  ? Iterators.<IEObjectDescription>emptyIterator() : resourceDescription.getExportedObjects(this.eClass).iterator();
		}
	}

}
