package org.openxma.dsl.dom.model;

import org.openxma.dsl.core.model.ModelElement;

/**
 * Strategy interface for determining various names, like column and table
 * names, given the respective model elements. This interface is used from the
 * default workflow and templates during a generator run.
 * 
 * Implementations use this to implement project-scoped naming standards.
 */
public interface NamingStrategy {
	/**
	 * Return a table name for an {@code Dao}
	 * 
	 * @param dao
	 *            the dao instance to generate the table name for
	 * @return a table name for the given dao
	 */
	String getTableName(Dao dao);

	/**
	 * Return the alias name for an {@code Dao}.
	 * 
	 * @param dao the dao instance to generate the table alias for
	 * @return an alias name for the given dao
	 */
	String getQualifier(Dao dao);

	/**
	 * Return the column name for an <code>Attribute</code>.
	 * 
	 * @param entity the entity of the given attribute 
	 * @param feature the feature to calculate the column for
	 * @return a column name for the given feature
	 */
	String getColumnName(Entity entity, Attribute feature);

	/**
	 * Return the column name for a nested (embeddable) <code>Attribute</code>.
	 * 
	 * @param entity the entity of the given attribute 
	 * @param feature the feature to calculate the column for
	 * @param nestedfeature
	 * @return a column name for the given nested feature
	 */
	String getColumnName(Entity entity, Attribute feature, Attribute nestedfeature);

	/**
	 * Return the implementation package name for the given modeElement.
	 * 
	 * @param modeElement
	 *            the modeElement instance to calculate the package for
	 * @return the implementation package name for the given modeElement
	 */
	String getPackageName(ModelElement modeElement);

	/**
	 * Return the interface package name for the given modeElement.
	 * 
	 * @param modeElement
	 *            the modeElement instance to calculate the package for
	 * @return the interface package name for the given modeElement
	 */
	String getInterfacePackageName(ModelElement modeElement);
	/**
	 * Provider that associates a <code>NamingStrategy</code> instance with the current thread
	 */
	class Holder {
		private static final ThreadLocal<NamingStrategy> namingStrategyHolder = new ThreadLocal<NamingStrategy>();

		/**
		 * Reset the NamingStrategy for the current thread.
		 */
		public static void resetNamingStrategy() {
			namingStrategyHolder.set(null);
		}

		/**
		 * Associate the given NamingStrategy with the current thread, <i>not</i>
		 * exposing it as inheritable for child threads.
		 * 
		 * @param modelModifier
		 *            the current NamingStrategy, or <code>null</code> to reset
		 *            the thread-bound context
		 */
		public static void setNamingStrategy(NamingStrategy modelModifier) {
			namingStrategyHolder.set(modelModifier);
		}

		/**
		 * Return the NamingStrategy associated with the current thread, if any.
		 * 
		 * @return the current NamingStrategy
		 */
		public static NamingStrategy getNamingStrategy() {
			return namingStrategyHolder.get();
		}
	}

}
