package org.openxma.dsl.dom.validation;

import static org.eclipse.emf.ecore.util.EcoreUtil.getObjectsByType;
import static org.eclipse.xtext.util.SimpleAttributeResolver.NAME_RESOLVER;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.xtext.EcoreUtil2;
import org.eclipse.xtext.util.SimpleAttributeResolver;
import org.eclipse.xtext.util.Strings;
import org.eclipse.xtext.validation.Check;
import org.eclipse.xtext.validation.CheckType;
import org.openxma.dsl.core.CorePackage;
import org.openxma.dsl.core.model.ModelElement;
import org.openxma.dsl.core.model.SimpleType;
import org.openxma.dsl.core.model.Style;
import org.openxma.dsl.core.model.Type;
import org.openxma.dsl.core.validation.CoreDslJavaValidator;
import org.openxma.dsl.dom.DomPackage;
import org.openxma.dsl.dom.model.Attribute;
import org.openxma.dsl.dom.model.ComplexType;
import org.openxma.dsl.dom.model.Dao;
import org.openxma.dsl.dom.model.DataView;
import org.openxma.dsl.dom.model.DelegateOperation;
import org.openxma.dsl.dom.model.Entity;
import org.openxma.dsl.dom.model.FeatureReference;
import org.openxma.dsl.dom.model.FromClass;
import org.openxma.dsl.dom.model.FromRange;
import org.openxma.dsl.dom.model.ManyToMany;
import org.openxma.dsl.dom.model.OneToOne;
import org.openxma.dsl.dom.model.Parameter;
import org.openxma.dsl.dom.model.QueryOperation;
import org.openxma.dsl.dom.model.QueryParameter;
import org.openxma.dsl.dom.model.QueryParameterValue;
import org.openxma.dsl.dom.model.SelectStatement;
import org.openxma.dsl.dom.model.ValueObject;
import org.openxma.dsl.dom.model.impl.SelectStatementImpl;

import com.google.common.base.Predicate;
import com.google.common.collect.Collections2;
import com.google.common.collect.ListMultimap;
import com.google.common.collect.Lists;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multimaps;
import com.google.common.collect.Sets;

public class DomDslJavaValidator extends CoreDslJavaValidator {
    public static final List<String> SUPPORTED_ID_TYPES = Lists.newArrayList("String", "Long", "Integer");
    public static final List<String> SUPPORTED_VERSION_TYPES = Lists.newArrayList("Date", "Long", "Timestamp");
    public static final String MISSING_IDENTIFIER = "DOM_MISSING_IDENTIFIER";
    public static final String INVALID_TYPE_NAME = "INVALID_TYPE_NAME";

    public static Set<String> RESERVED_KEYWORDS = Sets.newHashSet("abstract", "default", "if", "private", "this",
            "boolean", "do", "implements", "protected", "throw", "break", "double", "import", "public", "throws",
            "byte", "else", "instanceof", "return", "transient", "case", "extends", "int", "short", "try", "catch",
            "final", "interface", "static", "void", "char", "finally", "long", "strictfp", "volatile", "class",
            "float", "native", "super", "while", "const", "for", "new", "switch", "continue", "goto", "package",
            "synchronized", "true", "false", "null");

    @Override
    protected List<EPackage> getEPackages() {
        List<EPackage> result = new ArrayList<EPackage>();
        result.add(org.openxma.dsl.core.CorePackage.eINSTANCE);
        result.add(org.openxma.dsl.dom.DomPackage.eINSTANCE);
        return result;
    }

    @Check(CheckType.NORMAL)
    public void checkUnusedQueryParameter(QueryOperation queryOperation) {
        if (queryOperation.getStatement() != null) {
            Collection<QueryParameter> declaredQueryParameters = Collections2.filter(queryOperation
                    .getQueryParameters(), new Predicate<QueryParameter>() {
                public boolean apply(QueryParameter queryParameter) {
                    if (queryParameter instanceof Parameter) {
                        Parameter parameter = (Parameter) queryParameter;
                        if ("firstResult".equalsIgnoreCase(parameter.getName())
                                || "maxResults".equalsIgnoreCase(parameter.getName())) {
                            return false;
                        }
                    }
                    return true;
                }
            });
            List<QueryParameterValue> queryParameterValues = EcoreUtil2.eAllOfType(queryOperation.getStatement(),
                    QueryParameterValue.class);
            if (queryParameterValues.size() < declaredQueryParameters.size()) {
                warning("Unused QueryParameter's.", DomPackage.QUERY_OPERATION__STATEMENT);
            }
        }
    }

    @Check(CheckType.NORMAL)
    public void checkType(QueryOperation queryOperation) {
        if (queryOperation.getType() != null && queryOperation.getStatement() != null
                && SelectStatementImpl.class.equals(queryOperation.getStatement().getClass())) {
            SelectStatement statement = (SelectStatement) queryOperation.getStatement();
            EList<FromRange> from = statement.getFrom();
            if (from.size() == 1 && from.iterator().next() instanceof FromClass) {
                FromClass fromClass = (FromClass) from.iterator().next();
                if (!queryOperation.getType().equals(fromClass.getEntity())) {
                    error("Type mismatch: cannot convert from '"
                            + SimpleAttributeResolver.NAME_RESOLVER.apply(fromClass.getEntity()) + "' to '"
                            + SimpleAttributeResolver.NAME_RESOLVER.apply(queryOperation.getType()) + "'.",
                            DomPackage.QUERY_OPERATION__TYPE);
                }
            }
        }
    }

    @Check(CheckType.NORMAL)
    public void checkFeatureName(Attribute feature) {
        if (feature.getName() != null && RESERVED_KEYWORDS.contains(feature.getName())) {
            error("Invalid feature name: Reserved java keyword.", DomPackage.ATTRIBUTE__NAME);
        }
        if (!Character.isLowerCase(feature.getName().charAt(0))) {
            error("Invalid feature name: Name should start with a lowercase initial letter.",
                    DomPackage.ATTRIBUTE__NAME);
        }
    }

    @Check(CheckType.NORMAL)
    public void checkModelElement(ModelElement modelElement) {
        if (!(modelElement instanceof Style) && !Character.isUpperCase(modelElement.getName().charAt(0))) {
            warning("Name should start with a capital letter.", CorePackage.MODEL_ELEMENT__NAME, INVALID_TYPE_NAME,
                    modelElement.getName());
        }
    }

    @Check(CheckType.NORMAL)
    public void checkDelegateOperationDataView(DelegateOperation delegateOperation) {
        Entity entity = delegateOperation.getRepository().getEntity();
        DataView viewParameter = delegateOperation.getViewParameter() != null ? delegateOperation.getViewParameter()
                : entity.getDefaultDataView();
        if (delegateOperation.isCreateOperation()) {
            if (null != viewParameter && !viewParameter.includesAllRequiredFeaturesFor(entity)) {
                warning("DataView '" + viewParameter.getName()
                        + "' can not be used for create because it doesn't contain all required features of entity '"
                        + entity.getName() + "'", DomPackage.DELEGATE_OPERATION__VIEW_PARAMETER);
            }
        }
        if (delegateOperation.isUpdateOperation()) {
            if (null != viewParameter && !viewParameter.includesIdentifierFor(entity)) {
                boolean includesKey = false;
                if (entity.getKey() != null && entity.getIdentifier() == null
                        && (entity.getSuperType() == null || entity.getSuperType().getIdentifier() == null)) {
                    // check key
                    includesKey = viewParameter.includesKeyFor(entity);
                }
                if (!includesKey) {
                    warning(
                            "DataView '"
                                    + viewParameter.getName()
                                    + "' can not be used for update because it doesn't include the identifier or key attributes of entity '"
                                    + entity.getName() + "'", DomPackage.DELEGATE_OPERATION__VIEW_PARAMETER);
                }
            }
            if (entity.getVersion() != null && !viewParameter.includesVersionFor(entity)) {
                warning("DataView '" + viewParameter.getName() + "' is missing the version attribute of entity '"
                        + entity.getName() + "'", DomPackage.DELEGATE_OPERATION__VIEW_PARAMETER);
            }
        }

        if (delegateOperation.getFilter() != null) {
            if (!(delegateOperation.getOperation() instanceof QueryOperation)) {
                error("Filter expression only supported for QueryOperations.", DomPackage.DELEGATE_OPERATION__OPERATION);
            } else {
                QueryOperation queryOperation = (QueryOperation) delegateOperation.getOperation();
                if (!(queryOperation.getStatement() instanceof SelectStatement)) {
                    error("Filter expression only supported for SelectStatements.",
                            DomPackage.DELEGATE_OPERATION__OPERATION);
                }
                if (!queryOperation.isMany() || queryOperation.getType() == null) {
                    error("Filter expression only supported for operations returning collections.",
                            DomPackage.DELEGATE_OPERATION__OPERATION);
                }
            }
        }
    }

    @Check(CheckType.NORMAL)
    public void checkSameResourceWithEntity(Dao dao) {
        if (!dao.eResource().equals(dao.getEntity().eResource())) {
            error("Dao and entity must be defined in the same resource (dsl file).", DomPackage.DAO__ENTITY);
        }
    }

    @Check(CheckType.NORMAL)
    public void checkManyToMany(ManyToMany manyToMany) {
        if (manyToMany.getAttribute().getOpposite() == null && manyToMany.getAttribute().getOppositeReference() == null) {
            warning("Missing opposite reference definition.", manyToMany.getAttribute(), DomPackage.ATTRIBUTE__OPPOSITE);
        }
    }

    @Check(CheckType.NORMAL)
    public void checkIdentifierOrKeyAvailable(Dao dao) {
        Entity entity = dao.getEntity();
        if (entity.getKey() == null && entity.getIdentifier() == null) {
            error("Entity '" + entity.getName()
                    + "' is not persistable because it doesn't define an identifier attribute or a business key.",
                    DomPackage.DAO__ENTITY, MISSING_IDENTIFIER, entity.getName() == null ? "n/a" : entity.getName());
        }
    }

    @Check(CheckType.NORMAL)
    public void checkEntity(Entity entity) {
        Attribute identifierAttribute = null;
        Attribute versionAttribute = null;
        for (Attribute attribute : entity.getAllAttributes()) {
            if (attribute.getAttribute().isIdentifier()) {
                if (identifierAttribute != null) {
                    error("Entity '" + entity.getName() + "' defines more than one identifier attribute.",
                            DomPackage.DAO__ENTITY);
                } else {
                    identifierAttribute = attribute;
                    if (identifierAttribute.getDataTypeName() != null
                            && !SUPPORTED_ID_TYPES.contains(identifierAttribute.getDataTypeName())) {
                        Type type = identifierAttribute.getDataType();
                        if (isScalarType(type)) {
                            error("Unsupported identifier datatype '" + identifierAttribute.getDataTypeName()
                                    + "'. Supported types are: " + Strings.concat(",", SUPPORTED_ID_TYPES) + ".",
                                    DomPackage.ATTRIBUTE__TYPE);
                        }
                    }
                }
            }
            if (attribute.getAttribute().isVersion()) {
                if (versionAttribute != null) {
                    error("Entity '" + entity.getName() + "' defines more than one version attribute.",
                            DomPackage.DAO__ENTITY);
                } else {
                    versionAttribute = attribute;
                    if (versionAttribute.getDataTypeName() != null
                            && !SUPPORTED_VERSION_TYPES.contains(versionAttribute.getDataTypeName())) {
                        error("Unsupported version datatype '" + versionAttribute.getDataTypeName()
                                + "'. Supported types are: " + Strings.concat(",", SUPPORTED_VERSION_TYPES) + ".",
                                DomPackage.ATTRIBUTE__TYPE);
                    }
                }
            }
            if (attribute.isComposition()) {
                if (attribute.getOpposite() == null) {
                    error("Opposite reference must not be null for composition associations.", attribute,
                            DomPackage.ATTRIBUTE);
                }
            }
        }

    }

    private boolean isScalarType(Type type) {
        return type instanceof SimpleType && ((SimpleType) type).getTypeDefinition() != null
                && ((SimpleType) type).getTypeDefinition().getInstanceType() != null;
    }

    @Check(CheckType.NORMAL)
    public void checkSuperTypeCycle(Entity entity) {
        if (entity.getSuperType() != null && entity.getSuperType().equals(entity)) {
            error("Cycle detected: Entity '" + entity.getName() + "' cannot extend itself.",
                    DomPackage.ENTITY__SUPER_TYPE);
        }
    }

    @Check(CheckType.NORMAL)
    public void checkOneToOne(OneToOne oneToOne) {
        if (oneToOne.getAttribute().getOpposite() == null) {
            error("Missing opposite reference definition.", DomPackage.ATTRIBUTE__OPPOSITE);
        }
    }

    @Check(CheckType.NORMAL)
    public void checkReference(Attribute attribute) {
        if (attribute.eContainer() instanceof Entity) {
            if (!attribute.isMany() && attribute.isComposition() && null == attribute.getOpposite()) {
                error("Containment type reference must define the opposite reference.",
                        DomPackage.ATTRIBUTE__COMPOSITION);
            }
            if (attribute.isMany() && !attribute.isComposition() && null != attribute.getOpposite()
                    && attribute.getOpposite().isRequired()) {
                warning("Should be marked as composition since reference '" + attribute.getName()
                        + "' is marked as required on the opposite site '" + attribute.getOpposite().getName() + "'.",
                        DomPackage.ATTRIBUTE__OPPOSITE);
            }
            if (attribute.getOpposite() == attribute) {
                error("Opposite reference must not be the reference itself.", DomPackage.ATTRIBUTE__OPPOSITE);
            }
        }
    }

    @Check(CheckType.NORMAL)
    public void checkDuplicateAttributes(DataView dataView) {
        Multimap<String, Attribute> featureMap = Multimaps.newArrayListMultimap();
        for (FeatureReference featureReference : dataView.getFeatureReferences()) {
            Entity entity = featureReference.getSource();
            String name = featureReference.getSource().getName();
            if (featureReference.isAll()) {
                for (Attribute attribute : cloneAttributes(entity)) {
                    if (!(attribute.getDataType() instanceof Entity)) {
                        featureMap.put(name + attribute.getName(), attribute);
                    }
                }
            } else if (null != featureReference.getAttribute()) {
                featureMap.put(name
                        + (featureReference.getName() != null ? featureReference.getName() : featureReference
                                .getAttribute().getName()), featureReference.getAttribute());
            }
        }
        for (String name : featureMap.keySet()) {
            Collection<Attribute> features = featureMap.get(name);
            if (features.size() > 1) {
                for (Attribute feature : features) {
                    error("Duplicate attribute '" + feature.getName() + "'", DomPackage.DATA_VIEW__ATTRIBUTES);
                }
            }
        }
    }

    @Check(CheckType.NORMAL)
    public void checkAttributeTypes(ComplexType complexType) {
        if (complexType instanceof ValueObject) {
            for (Attribute attribute : complexType.getAttributes()) {
                if (attribute.getDataType() instanceof Entity) {
                    error("Entity references not supported in '" + complexType.eClass().getName() + "' elements",
                            attribute, DomPackage.ATTRIBUTE__DATA_TYPE);
                }
            }
        }
    }

    @Check(CheckType.NORMAL)
    public void checkIfAttributeIsUnique(final Attribute attribute) {
        Collection<EObject> attributes = getObjectsByType(attribute.eContainer().eContents(), DomPackage.eINSTANCE
                .getAttribute());
        final ListMultimap<String, ? extends EObject> multiMap = Multimaps.index(attributes, NAME_RESOLVER);
        Collection<EObject> duplicateAttributeNames = Collections2.filter(attributes, new Predicate<EObject>() {
            public boolean apply(EObject input) {
                return input.equals(attribute) && multiMap.get(NAME_RESOLVER.apply(input)).size() > 1;
            }
        });
        if (!duplicateAttributeNames.isEmpty()) {
            error("Duplicate attribute '" + attribute.getName() + "'", DomPackage.IELEMENT_WITH_NO_NAME__NO_NAME);
        }
    }

    /**
     * Returns a collection of the self-contained copies of each attribute.
     * 
     * @param entity
     *            the collection of attribute to copy.
     * @return the collection of attribute.
     */
    private static Collection<Attribute> cloneAttributes(Entity entity) {
        Collection<Attribute> attributes = EcoreUtil.copyAll(entity.getAttributes());
        if (entity.getSuperType() != null) {
            Collection<Attribute> superTypeAttributes = cloneAttributes(entity.getSuperType());
            attributes.addAll(EcoreUtil.copyAll(superTypeAttributes));
        }
        return attributes;
    }

}
